#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import os
import idaes.logger as idaeslog
from pyomo.environ import (
    ConcreteModel,
    Expression,
    value,
    TransformationFactory,
    units as pyunits,
)
from pyomo.network import Arc, SequentialDecomposition
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.models.unit_models import Product
import idaes.core.util.scaling as iscale
from idaes.core import UnitModelCostingBlock

from watertap.core.util.initialization import assert_degrees_of_freedom, check_solve

from watertap.core.wt_database import Database
import watertap.core.zero_order_properties as prop_ZO
from watertap.unit_models.zero_order import (
    FeedZO,
    ATHTLZO,
    SaltPrecipitationZO,
    HTGZO,
)
from watertap.costing.zero_order_costing import ZeroOrderCosting

# Set up logger
_log = idaeslog.getLogger(__name__)


def main():
    m = build()

    set_operating_conditions(m)
    assert_degrees_of_freedom(m, 0)
    assert_units_consistent(m)

    initialize_system(m)

    results = solve(m, checkpoint="solve flowsheet after initializing system")
    display_results(m)

    add_costing(m)
    initialize_costing(m)
    assert_degrees_of_freedom(m, 0)
    assert_units_consistent(m)

    results = solve(m, checkpoint="solve flowsheet after costing")

    display_metrics_results(m)
    display_additional_results(m)

    return m, results


def build():
    # flowsheet set up
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.prop = prop_ZO.WaterParameterBlock(
        solute_list=[
            "organic_solid",
            "organic_liquid",
            "inorganic_solid",
            "carbon_dioxide",
        ]
    )

    # unit models
    m.fs.feed = FeedZO(property_package=m.fs.prop)
    m.fs.ATHTL = ATHTLZO(property_package=m.fs.prop, database=m.db)
    m.fs.salt_precipitation = SaltPrecipitationZO(
        property_package=m.fs.prop, database=m.db
    )
    m.fs.HTG = HTGZO(property_package=m.fs.prop, database=m.db)

    m.fs.product_H2O = Product(property_package=m.fs.prop)
    # CO2 from ATHTL to sulfer conversion unit, final product should be solid sulfer, H2 and CO2
    m.fs.product_CO2 = Product(property_package=m.fs.prop)
    # sulfates, nitrates & phosphates (organics & inorganics) from salt precipitation unit
    m.fs.product_salts = Product(property_package=m.fs.prop)
    # CO2 from HTG for renewable natural gas product
    m.fs.product_natural_gas = Product(property_package=m.fs.prop)

    # connections
    m.fs.s01 = Arc(source=m.fs.feed.outlet, destination=m.fs.ATHTL.inlet)
    m.fs.s02 = Arc(source=m.fs.ATHTL.treated, destination=m.fs.salt_precipitation.inlet)
    m.fs.s03 = Arc(source=m.fs.salt_precipitation.treated, destination=m.fs.HTG.inlet)
    m.fs.s04 = Arc(source=m.fs.HTG.treated, destination=m.fs.product_H2O.inlet)
    m.fs.s05 = Arc(
        source=m.fs.HTG.byproduct, destination=m.fs.product_natural_gas.inlet
    )
    m.fs.s06 = Arc(source=m.fs.ATHTL.byproduct, destination=m.fs.product_CO2.inlet)
    m.fs.s07 = Arc(
        source=m.fs.salt_precipitation.byproduct, destination=m.fs.product_salts.inlet
    )
    TransformationFactory("network.expand_arcs").apply_to(m)

    # scaling
    iscale.calculate_scaling_factors(m)

    return m


def set_operating_conditions(m):
    # ---specifications---
    # feed
    flow_mass_H2O = 400 * pyunits.metric_ton / pyunits.day
    flow_mass_inorganic_solid = 28.8 * pyunits.metric_ton / pyunits.day
    flow_mass_organic_solid = 71.2 * pyunits.metric_ton / pyunits.day

    m.fs.feed.flow_mass_comp[0, "H2O"].fix(flow_mass_H2O)
    m.fs.feed.flow_mass_comp[0, "inorganic_solid"].fix(flow_mass_inorganic_solid)
    m.fs.feed.flow_mass_comp[0, "organic_solid"].fix(flow_mass_organic_solid)
    m.fs.feed.conc_mass_comp[0, "organic_liquid"].fix(1e-5)
    m.fs.feed.conc_mass_comp[0, "carbon_dioxide"].fix(1e-5)
    solve(m.fs.feed, checkpoint="solve feed block")

    # ATHTL
    m.fs.ATHTL.load_parameters_from_database(use_default_removal=True)

    # Salt precipitation
    m.fs.salt_precipitation.load_parameters_from_database(use_default_removal=True)

    # HTG
    m.fs.HTG.load_parameters_from_database(use_default_removal=True)


def initialize_system(m):
    seq = SequentialDecomposition()
    seq.options.tear_set = []
    seq.options.iterLim = 1
    seq.run(m, lambda u: u.initialize())


def solve(blk, solver=None, checkpoint=None, tee=False, fail_flag=True):
    if solver is None:
        solver = get_solver()
    results = solver.solve(blk, tee=tee)
    check_solve(results, checkpoint=checkpoint, logger=_log, fail_flag=fail_flag)
    return results


def display_results(m):
    unit_list = ["feed", "ATHTL", "salt_precipitation", "HTG"]
    for u in unit_list:
        m.fs.component(u).report()


def add_costing(m):
    source_file = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "supercritical_sludge_to_gas_global_costing.yaml",
    )
    m.fs.costing = ZeroOrderCosting(case_study_definition=source_file)
    # typing aid
    costing_kwargs = {"flowsheet_costing_block": m.fs.costing}
    m.fs.ATHTL.costing = UnitModelCostingBlock(**costing_kwargs)
    m.fs.salt_precipitation.costing = UnitModelCostingBlock(**costing_kwargs)
    m.fs.HTG.costing = UnitModelCostingBlock(**costing_kwargs)

    m.fs.costing.cost_process()
    m.fs.costing.add_electricity_intensity(m.fs.product_H2O.properties[0].flow_vol)
    m.fs.costing.add_LCOW(m.fs.product_H2O.properties[0].flow_vol)

    m.fs.costing.LCOG = Expression(
        expr=(
            m.fs.costing.total_capital_cost * m.fs.costing.capital_recovery_factor
            + m.fs.costing.total_operating_cost
        )
        / (
            m.fs.costing.utilization_factor
            * pyunits.convert(
                m.fs.product_natural_gas.properties[0].flow_mass_comp["carbon_dioxide"],
                to_units=pyunits.kg / m.fs.costing.base_period,
            )
        ),
        doc="Levelized cost of CO2 for production of renewable natural gas",
    )

    m.fs.costing.LCOC = Expression(
        expr=(
            m.fs.costing.total_capital_cost * m.fs.costing.capital_recovery_factor
            + m.fs.costing.total_operating_cost
        )
        / (
            m.fs.costing.utilization_factor
            * pyunits.convert(
                m.fs.product_CO2.properties[0].flow_mass_comp["carbon_dioxide"],
                to_units=pyunits.kg / m.fs.costing.base_period,
            )
        ),
        doc="Levelized cost of CO2 from AT-HTL",
    )

    m.fs.costing.LCOS = Expression(
        expr=(
            m.fs.costing.total_capital_cost * m.fs.costing.capital_recovery_factor
            + m.fs.costing.total_operating_cost
        )
        / (
            m.fs.costing.utilization_factor
            * pyunits.convert(
                (
                    m.fs.product_salts.properties[0].flow_mass_comp["organic_solid"]
                    + m.fs.product_salts.properties[0].flow_mass_comp["inorganic_solid"]
                ),
                to_units=pyunits.kg / m.fs.costing.base_period,
            )
        ),
        doc="Levelized cost of Sulfates, Nitrates and Phosphates",
    )

    # other levelized costs
    m.fs.costing.annual_water_inlet = Expression(
        expr=m.fs.costing.utilization_factor
        * pyunits.convert(
            m.fs.feed.properties[0].flow_vol,
            to_units=pyunits.m**3 / m.fs.costing.base_period,
        )
    )

    m.fs.costing.annual_water_production = Expression(
        expr=m.fs.costing.utilization_factor
        * pyunits.convert(
            m.fs.product_H2O.properties[0].flow_vol,
            to_units=pyunits.m**3 / m.fs.costing.base_period,
        )
    )

    m.fs.costing.annual_CO2_renewable_production = Expression(
        expr=(
            m.fs.costing.utilization_factor
            * pyunits.convert(
                m.fs.HTG.byproduct.flow_mass_comp[0, "carbon_dioxide"],
                to_units=pyunits.kg / m.fs.costing.base_period,
            )
        )
    )

    m.fs.costing.annual_CO2_sulfur_production = Expression(
        expr=(
            m.fs.costing.utilization_factor
            * pyunits.convert(
                m.fs.ATHTL.byproduct.flow_mass_comp[0, "carbon_dioxide"],
                to_units=pyunits.kg / m.fs.costing.base_period,
            )
        )
    )

    m.fs.costing.annual_sulphate_nitrate_phosphate_production = Expression(
        expr=(
            m.fs.costing.utilization_factor
            * pyunits.convert(
                (
                    m.fs.salt_precipitation.byproduct.flow_mass_comp[0, "organic_solid"]
                    + m.fs.salt_precipitation.byproduct.flow_mass_comp[
                        0, "inorganic_solid"
                    ]
                ),
                to_units=pyunits.kg / m.fs.costing.base_period,
            )
        )
    )

    m.fs.costing.total_annualized_cost = Expression(
        expr=(
            m.fs.costing.total_capital_cost * m.fs.costing.capital_recovery_factor
            + m.fs.costing.total_operating_cost
        )
    )

    m.fs.costing.LCOT = Expression(
        expr=(m.fs.costing.total_annualized_cost / m.fs.costing.annual_water_inlet),
        doc="Levelized Cost of Treatment",
    )


def initialize_costing(m):
    m.fs.costing.initialize()


def display_costing(m):
    m.fs.costing.total_capital_cost.display()
    m.fs.costing.total_operating_cost.display()
    m.fs.costing.LCOW.display()

    print("\nUnit Capital Costs\n")
    for u in m.fs.costing._registered_unit_costing:
        print(
            u.name,
            " :   ",
            value(pyunits.convert(u.capital_cost, to_units=pyunits.USD_2020)),
        )

    print("\nUtility Costs\n")
    for f in m.fs.costing.used_flows:
        print(
            f,
            " :   ",
            value(
                pyunits.convert(
                    m.fs.costing.aggregate_flow_costs[f],
                    to_units=pyunits.USD_2020 / pyunits.year,
                )
            ),
        )

    print("")
    total_capital_cost = value(
        pyunits.convert(m.fs.costing.total_capital_cost, to_units=pyunits.MUSD_2020)
    )
    print(f"Total Capital Costs: {total_capital_cost:.2f} M$")
    total_operating_cost = value(
        pyunits.convert(
            m.fs.costing.total_operating_cost, to_units=pyunits.MUSD_2020 / pyunits.year
        )
    )
    print(f"Total Operating Costs: {total_operating_cost:.2f} M$/year")
    electricity_intensity = value(
        pyunits.convert(
            m.fs.costing.electricity_intensity, to_units=pyunits.kWh / pyunits.m**3
        )
    )
    print(f"Electricity Intensity: {electricity_intensity:.4f} kWh/m^3")
    LCOW = value(
        pyunits.convert(m.fs.costing.LCOW, to_units=pyunits.USD_2020 / pyunits.m**3)
    )
    print(f"Levelized Cost of Water: {LCOW:.4f} $/m^3")
    LCOG = value(
        pyunits.convert(m.fs.costing.LCOG, to_units=pyunits.USD_2020 / pyunits.kg)
    )
    print(f"Levelized Cost of Natural Gas Based on Carbon Dioxide: {LCOG:.3f} $/kg")
    LCOC = value(
        pyunits.convert(m.fs.costing.LCOC, to_units=pyunits.USD_2020 / pyunits.kg)
    )
    print(f"Levelized Cost of Carbon Dioxide from AT-HTL: {LCOC:.3f} $/kg")
    LCOS = value(
        pyunits.convert(m.fs.costing.LCOS, to_units=pyunits.USD_2020 / pyunits.kg)
    )
    print(f"Levelized Cost of Sulfates, Nitrates and Phosphates: {LCOS:.3f} $/kg")


def display_metrics_results(m):
    print("----------Levelized costs----------")
    LCOT = value(
        pyunits.convert(
            m.fs.costing.LCOT, to_units=m.fs.costing.base_currency / pyunits.m**3
        )
    )
    print(f"Levelized Cost of Treatment: {LCOT:.4f} $/m3 of feed")
    LCOW = value(
        pyunits.convert(
            m.fs.costing.LCOW, to_units=m.fs.costing.base_currency / pyunits.m**3
        )
    )
    print(f"Levelized Cost of Water: {LCOW:.4f} $/m3 of product")
    LCOG = value(
        pyunits.convert(m.fs.costing.LCOG, to_units=pyunits.USD_2020 / pyunits.kg)
    )
    print(f"Levelized Cost of Natural Gas Based on Carbon Dioxide: {LCOG:.4f} $/kg")
    LCOC = value(
        pyunits.convert(m.fs.costing.LCOC, to_units=pyunits.USD_2020 / pyunits.kg)
    )
    print(f"Levelized Cost of Carbon Dioxide from AT-HTL: {LCOC:.4f} $/kg")
    LCOS = value(
        pyunits.convert(m.fs.costing.LCOS, to_units=pyunits.USD_2020 / pyunits.kg)
    )
    print(f"Levelized Cost of Sulfates, Nitrates and Phosphates: {LCOS:.4f} $/kg")

    print("----------Capital costs----------")
    DCC_normalized = value(
        pyunits.convert(
            (
                m.fs.ATHTL.costing.direct_capital_cost
                + m.fs.salt_precipitation.costing.direct_capital_cost
                + m.fs.HTG.costing.direct_capital_cost
            )
            / m.fs.feed.properties[0].flow_vol,
            to_units=m.fs.costing.base_currency / (pyunits.m**3 / pyunits.day),
        )
    )
    print(f"Normalized direct capital costs: {DCC_normalized:.4f} $/(m3/day)")
    ICC_normalized = value(
        pyunits.convert(
            m.fs.costing.total_capital_cost / m.fs.feed.properties[0].flow_vol,
            to_units=m.fs.costing.base_currency / (pyunits.m**3 / pyunits.day),
        )
    )
    print(f"Normalized total capital costs: {ICC_normalized:.4f} $/(m3/day)")

    print("----------Operating costs----------")
    FMC_normalized = value(
        pyunits.convert(
            m.fs.costing.maintenance_cost / m.fs.costing.total_capital_cost,
            to_units=1 / pyunits.a,
        )
    )
    print(f"Normalized maintenance costs: {FMC_normalized:.4f} 1/year")
    EC_normalized = value(
        pyunits.convert(
            m.fs.costing.aggregate_flow_costs["electricity"]
            / m.fs.costing.annual_water_inlet,
            to_units=m.fs.costing.base_currency / pyunits.m**3,
        )
    )
    print(f"Normalized electricity cost: {EC_normalized:.4f} $/m3 of feed")
    Catalyst_ATHTL_normalized = value(
        pyunits.convert(
            m.fs.costing.aggregate_flow_costs["catalyst_ATHTL"]
            / m.fs.costing.annual_water_inlet,
            to_units=m.fs.costing.base_currency / pyunits.m**3,
        )
    )
    print(
        f"Normalized catalyst for AT-HTL cost: {Catalyst_ATHTL_normalized:.4f} $/m3 of feed"
    )
    Catalyst_HTG_normalized = value(
        pyunits.convert(
            m.fs.costing.aggregate_flow_costs["catalyst_HTG"]
            / m.fs.costing.annual_water_inlet,
            to_units=m.fs.costing.base_currency / pyunits.m**3,
        )
    )
    print(
        f"Normalized catalyst for HTG cost: {Catalyst_HTG_normalized:.4f} $/m3 of feed"
    )

    print("----------Performance metrics----------")
    volumetric_recovery = value(
        m.fs.product_H2O.properties[0].flow_vol / m.fs.feed.properties[0].flow_vol
    )
    print(f"Water recovery: {volumetric_recovery:.4f} m3 of product/m3 of feed")
    OrganicR_normalized = value(
        pyunits.convert(
            1
            - (
                m.fs.product_H2O.properties[0].flow_mass_comp["organic_solid"]
                + m.fs.product_H2O.properties[0].flow_mass_comp["organic_liquid"]
            )
            / (
                m.fs.feed.properties[0].flow_mass_comp["organic_solid"]
                + m.fs.feed.properties[0].flow_mass_comp["organic_liquid"]
            ),
            to_units=pyunits.dimensionless,
        )
    )
    print(f"Organic removal: {OrganicR_normalized:.4f} dimensionless")
    InorganicR_normalized = value(
        pyunits.convert(
            1
            - m.fs.product_H2O.properties[0].flow_mass_comp["inorganic_solid"]
            / m.fs.feed.properties[0].flow_mass_comp["inorganic_solid"],
            to_units=pyunits.dimensionless,
        )
    )
    print(f"Inorganic removal: {InorganicR_normalized:.4f} dimensionless")

    print("----------Energy intensity----------")
    SEC = value(
        pyunits.convert(
            m.fs.costing.aggregate_flow_electricity / m.fs.feed.properties[0].flow_vol,
            to_units=pyunits.kWh / pyunits.m**3,
        )
    )
    print(f"Specific electricity consumption: {SEC:.4f} kWh/m3 of feed")


def display_additional_results(m):
    print("----------Outlets----------")
    product_H2O_flow = value(
        pyunits.convert(
            m.fs.product_H2O.properties[0].flow_vol,
            to_units=pyunits.m**3 / pyunits.hr,
        )
    )
    print(f"H2O outlet flow: {product_H2O_flow:.4f} m3/h")
    product_H2O_Organic_Solid = value(
        pyunits.convert(
            m.fs.product_H2O.properties[0].conc_mass_comp["organic_solid"],
            to_units=pyunits.g / pyunits.L,
        )
    )
    print(f"H2O outlet organic(s) conc: {product_H2O_Organic_Solid:.4f} g/L")
    product_H2O_Organic_Liquid = value(
        pyunits.convert(
            m.fs.product_H2O.properties[0].conc_mass_comp["organic_liquid"],
            to_units=pyunits.g / pyunits.L,
        )
    )
    print(f"H2O outlet Organic(l) conc: {product_H2O_Organic_Liquid:.4f} g/L")
    product_H2O_Inorganic_Solid = value(
        pyunits.convert(
            m.fs.product_H2O.properties[0].conc_mass_comp["inorganic_solid"],
            to_units=pyunits.g / pyunits.L,
        )
    )
    print(f"H2O outlet inorganic(s) conc: {product_H2O_Inorganic_Solid:.4f} g/L")

    print("----------Capital costs----------")
    total_capital_costs = value(m.fs.costing.total_capital_cost) / 1e6
    print(f"Total capital costs: {total_capital_costs:.4f} $M")
    ATHTL_capital_costs = value(m.fs.ATHTL.costing.capital_cost) / 1e6
    print(f"AT-HTL capital costs: {ATHTL_capital_costs:.4f} $M")
    salt_precipitation_capital_costs = (
        value(m.fs.salt_precipitation.costing.capital_cost) / 1e6
    )
    print(
        f"Salt precipitation capital costs: {salt_precipitation_capital_costs:.4f} $M"
    )
    HTG_capital_costs = value(m.fs.HTG.costing.capital_cost) / 1e6
    print(f"HTG capital costs: {HTG_capital_costs:.4f} $M")

    print("----------Operating costs----------")
    total_operating_costs = value(m.fs.costing.total_operating_cost) / 1e6
    print(f"Total operating costs: {total_operating_costs:.4f} $M/year")
    fixed_operating_costs = value(m.fs.costing.total_fixed_operating_cost) / 1e6
    print(f"Fixed operating costs: {fixed_operating_costs:.4f} $M/year")
    electricity_operating_costs = (
        value(m.fs.costing.aggregate_flow_costs["electricity"]) / 1e3
    )
    print(f"Electricity operating costs: {electricity_operating_costs:.4f} $k/year")
    catalyst_ATHTL_operating_costs = (
        value(m.fs.costing.aggregate_flow_costs["catalyst_ATHTL"]) / 1e6
    )
    print(
        f"Catalyst in AT-HTL operating costs: {catalyst_ATHTL_operating_costs:.4f} $M/year"
    )
    catalyst_HTG_operating_costs = (
        value(m.fs.costing.aggregate_flow_costs["catalyst_HTG"]) / 1e6
    )
    print(
        f"Catalyst in HTG operating costs: {catalyst_HTG_operating_costs:.4f} $M/year"
    )


if __name__ == "__main__":
    m, results = main()



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest

from pyomo.util.check_units import assert_units_consistent
from pyomo.environ import (
    value,
    assert_optimal_termination,
)

from watertap.core.util.initialization import assert_degrees_of_freedom

from watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.amo_1690.amo_1690 import (
    build,
    set_operating_conditions,
    initialize_system,
    solve,
    add_costing,
    display_costing,
    display_results,
)


class TestAMO1690Flowsheet:
    @pytest.fixture(scope="class")
    def system_frame(self):
        m = build()
        return m

    @pytest.mark.unit
    def test_build(self, system_frame):
        m = system_frame
        assert_degrees_of_freedom(m, 29)
        assert_units_consistent(m)

    @pytest.mark.component
    def test_set_operating_conditions(self, system_frame):
        m = system_frame
        set_operating_conditions(m)

        # test feed composition
        assert pytest.approx(1093.93, rel=1e-3) == value(
            m.fs.feed.flow_mass_comp[0, "H2O"]
        )
        assert pytest.approx(0.547658, rel=1e-3) == value(
            m.fs.feed.flow_mass_comp[0, "tss"]
        )
        assert pytest.approx(0.766721, rel=1e-3) == value(
            m.fs.feed.flow_mass_comp[0, "cod"]
        )
        assert pytest.approx(0.0711955, rel=1e-3) == value(
            m.fs.feed.flow_mass_comp[0, "tkn"]
        )
        assert value(m.fs.feed.flow_mass_comp[0, "acetic_acid"]) < 1e-8
        assert value(m.fs.feed.flow_mass_comp[0, "ammonium_as_nitrogen"]) < 1e-8

    @pytest.mark.component
    def test_initialize(self, system_frame):
        m = system_frame
        initialize_system(m)

    @pytest.mark.component
    def test_solve(self, system_frame):
        m = system_frame
        results = solve(m)
        assert_optimal_termination(results)

        # check filtered water
        assert value(
            m.fs.filtered_water.properties[0].flow_mass_comp["H2O"]
        ) == pytest.approx(984.537, rel=1e-3)
        assert value(
            m.fs.filtered_water.properties[0].flow_mass_comp["tss"]
        ) == pytest.approx(0.10953, rel=1e-3)
        assert value(
            m.fs.filtered_water.properties[0].flow_mass_comp["cod"]
        ) == pytest.approx(0.38336, rel=1e-3)
        assert value(
            m.fs.filtered_water.properties[0].flow_mass_comp["tkn"]
        ) == pytest.approx(0.0605162, rel=1e-3)
        assert (
            value(m.fs.filtered_water.properties[0].flow_mass_comp["acetic_acid"])
            < 1e-8
        )
        assert (
            value(
                m.fs.filtered_water.properties[0].flow_mass_comp["ammonium_as_nitrogen"]
            )
            < 1e-8
        )

        # check anaerobic digester byproduct and biogas
        assert value(m.fs.ad_byproduct.properties[0].flow_mass_comp["H2O"]) < 1e-8
        assert value(
            m.fs.ad_byproduct.properties[0].flow_mass_comp["tss"]
        ) == pytest.approx(0.21906, rel=1e-3)
        assert value(
            m.fs.ad_byproduct.properties[0].flow_mass_comp["cod"]
        ) == pytest.approx(0.19168, rel=1e-3)
        assert value(
            m.fs.ad_byproduct.properties[0].flow_mass_comp["tkn"]
        ) == pytest.approx(0.005339665, rel=1e-3)
        assert (
            value(m.fs.ad_byproduct.properties[0].flow_mass_comp["acetic_acid"]) < 1e-8
        )
        assert (
            value(
                m.fs.ad_byproduct.properties[0].flow_mass_comp["ammonium_as_nitrogen"]
            )
            < 1e-8
        )
        assert value(m.fs.ad.biogas_production[0]) == pytest.approx(0.17525, rel=1e-3)

        # check membrane evaporator treated
        assert value(
            m.fs.me_treated.properties[0].flow_mass_comp["H2O"]
        ) == pytest.approx(106.6582, rel=1e-3)
        assert value(m.fs.me_treated.properties[0].flow_mass_comp["tss"]) < 1e-8
        assert value(m.fs.me_treated.properties[0].flow_mass_comp["cod"]) < 1e-8
        assert value(m.fs.me_treated.properties[0].flow_mass_comp["tkn"]) < 1e-8
        assert value(
            m.fs.me_treated.properties[0].flow_mass_comp["acetic_acid"]
        ) == pytest.approx(0.00438126, rel=1e-3)
        assert value(
            m.fs.me_treated.properties[0].flow_mass_comp["ammonium_as_nitrogen"]
        ) == pytest.approx(0.01139129, rel=1e-3)

        # check membrane evaporator byproduct
        assert value(
            m.fs.me_byproduct.properties[0].flow_mass_comp["H2O"]
        ) == pytest.approx(2.73486, rel=1e-3)
        assert value(m.fs.me_byproduct.properties[0].flow_mass_comp["tss"]) < 1e-8
        assert value(m.fs.me_byproduct.properties[0].flow_mass_comp["cod"]) < 1e-8
        assert value(m.fs.me_byproduct.properties[0].flow_mass_comp["tkn"]) < 1e-8
        assert value(
            m.fs.me_byproduct.properties[0].flow_mass_comp["acetic_acid"]
        ) == pytest.approx(0.017525, rel=1e-3)
        assert value(
            m.fs.me_byproduct.properties[0].flow_mass_comp["ammonium_as_nitrogen"]
        ) == pytest.approx(0.045565, rel=1e-3)

    @pytest.mark.component
    def test_costing(self, system_frame):
        m = system_frame

        add_costing(m)
        m.fs.costing.initialize()
        results = solve(m)
        assert_optimal_termination(results)

        assert value(m.fs.costing.LCOT) == pytest.approx(0.133052087, rel=1e-3)
        assert value(m.fs.costing.LCOT_with_revenue) == pytest.approx(
            0.040252087, rel=1e-3
        )
        assert value(m.fs.costing.LC_biogas) == pytest.approx(0.8315755, rel=1e-3)
        assert value(m.fs.costing.LC_biogas_with_revenue) == pytest.approx(
            0.6515755, rel=1e-3
        )
        assert value(m.fs.costing.LC_fertilizer) == pytest.approx(2.309932, rel=1e-3)
        assert value(m.fs.costing.LC_fertilizer_with_revenue) == pytest.approx(
            1.1988209569997055, rel=1e-3
        )

    @pytest.mark.component
    def test_display(self, system_frame):
        m = system_frame
        display_results(m)
        display_costing(m)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for zero-order chemical addition model
"""
import pytest


from pyomo.environ import Block, ConcreteModel, Constraint, Param, value, Var
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.core.util.exceptions import ConfigurationError
from idaes.core import UnitModelCostingBlock

from watertap.unit_models.zero_order import ChemicalAdditionZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.costing.zero_order_costing import ZeroOrderCosting

solver = get_solver()


@pytest.mark.unit
def test_no_subtype():
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.params = WaterParameterBlock(solute_list=["sulfur", "toc", "tss"])

    with pytest.raises(
        ConfigurationError,
        match="fs.unit - zero-order chemical addition "
        "operations require the process_subtype configuration "
        "argument to be set",
    ):
        m.fs.unit = ChemicalAdditionZO(property_package=m.fs.params, database=m.db)


class TestChemAddZOAmmonia:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(solute_list=["sulfur", "toc", "tss"])

        m.fs.unit = ChemicalAdditionZO(
            property_package=m.fs.params, database=m.db, process_subtype="default"
        )

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(1000)
        m.fs.unit.inlet.flow_mass_comp[0, "sulfur"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "toc"].fix(2)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(3)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db

        assert isinstance(model.fs.unit.chemical_dosage, Var)
        assert isinstance(model.fs.unit.chemical_flow_vol, Var)
        assert isinstance(model.fs.unit.solution_density, Var)
        assert isinstance(model.fs.unit.ratio_in_solution, Var)
        assert isinstance(model.fs.unit.chemical_flow_constraint, Constraint)

        assert isinstance(model.fs.unit.lift_height, Param)
        assert isinstance(model.fs.unit.eta_pump, Param)
        assert isinstance(model.fs.unit.eta_motor, Param)
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.electricity_consumption, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("chemical_addition")

        model.fs.unit.load_parameters_from_database()

        assert model.fs.unit.chemical_dosage[0].fixed
        assert model.fs.unit.chemical_dosage[0].value == 1

        assert model.fs.unit.solution_density.fixed
        assert model.fs.unit.solution_density.value == 1000

        assert model.fs.unit.ratio_in_solution.fixed
        assert model.fs.unit.ratio_in_solution.value == 0.5

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        for t, j in model.fs.unit.inlet.flow_mass_comp:
            assert pytest.approx(
                value(model.fs.unit.inlet.flow_mass_comp[t, j]), rel=1e-5
            ) == value(model.fs.unit.outlet.flow_mass_comp[t, j])

        assert pytest.approx(2.012e-6, rel=1e-5) == value(
            model.fs.unit.chemical_flow_vol[0]
        )

        assert pytest.approx(7.41395e-4, rel=1e-5) == value(
            model.fs.unit.electricity[0]
        )

    @pytest.mark.component
    def test_report(self, model):

        model.fs.unit.report()


db = Database()
params = db._get_technology("chemical_addition")


class TestPumpZOsubtype:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(solute_list=["sulfur", "toc", "tss"])

        return m

    @pytest.mark.parametrize("subtype", [k for k in params.keys()])
    @pytest.mark.component
    def test_load_parameters(self, model, subtype):
        model.fs.unit = ChemicalAdditionZO(
            property_package=model.fs.params, database=db, process_subtype=subtype
        )

        model.fs.unit.config.process_subtype = subtype
        data = db.get_unit_operation_parameters("chemical_addition", subtype=subtype)

        model.fs.unit.load_parameters_from_database()

        assert model.fs.unit.chemical_dosage[0].fixed
        assert (
            model.fs.unit.chemical_dosage[0].value == data["chemical_dosage"]["value"]
        )

        assert model.fs.unit.solution_density.fixed
        assert model.fs.unit.solution_density.value == data["solution_density"]["value"]

        assert model.fs.unit.ratio_in_solution.fixed
        assert (
            model.fs.unit.ratio_in_solution.value == data["ratio_in_solution"]["value"]
        )


@pytest.mark.parametrize("subtype", [k for k in params.keys() if k != "default"])
def test_costing(subtype):
    print(subtype)
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = WaterParameterBlock(solute_list=["sulfur", "toc", "tss"])

    m.fs.costing = ZeroOrderCosting()

    m.fs.unit1 = ChemicalAdditionZO(
        property_package=m.fs.params, database=m.db, process_subtype=subtype
    )

    m.fs.unit1.inlet.flow_mass_comp[0, "H2O"].fix(10000)
    m.fs.unit1.inlet.flow_mass_comp[0, "sulfur"].fix(1)
    m.fs.unit1.inlet.flow_mass_comp[0, "toc"].fix(2)
    m.fs.unit1.inlet.flow_mass_comp[0, "tss"].fix(3)
    m.fs.unit1.load_parameters_from_database(use_default_removal=True)
    assert degrees_of_freedom(m.fs.unit1) == 0

    m.fs.unit1.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    assert isinstance(m.fs.costing.chemical_addition, Block)
    assert isinstance(m.fs.costing.chemical_addition.capital_a_parameter, Var)
    assert isinstance(m.fs.costing.chemical_addition.capital_b_parameter, Var)

    assert isinstance(m.fs.unit1.costing.capital_cost, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost_constraint, Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit1) == 0

    assert m.fs.unit1.electricity[0] in m.fs.costing._registered_flows["electricity"]

    assert str(
        m.fs.unit1.chemical_dosage[0]
        * m.fs.unit1.properties[0].flow_vol
        / m.fs.unit1.ratio_in_solution
    ) == str(m.fs.costing._registered_flows[subtype][0])



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
# WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING
#
# This module is a work in progress. Do not use it for real work right now.
#
# WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING

"""
JSON schema embedded as variables for:
  - component
  - reaction
"""

from .data_model import Reaction


_parameter_def = {
    "type": "array",
    "description": "List of parameter values",
    "items": {
        "type": "object",
        "description": "Value, units, etc. for a parameter",
        "properties": {
            "v": {"description": "value", "type": "number"},
            "u": {"description": "units", "type": "string"},
            "i": {
                "oneOf": [
                    {"type": "string", "description": "string index"},
                    {"type": "number", "description": "numeric index"},
                ]
            },
        },
        "required": ["v", "u"],
    },
}

schemas = {
    "component": {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "$id": "https://nawi-hub.github.com/electrolytedb/component",
        "type": "object",
        "description": "A chemical species that is a component in a reaction",
        "properties": {
            "name": {
                "description": "The chemical name of the component",
                "examples": ["HPO4 2-", "H3PO4", "NH4 +"],
                "type": "string",
            },
            "elements": {
                "description": "List of elements",
                "type": "array",
                "items": {
                    "type": "string",
                    "description": "Name of an individual element",
                },
            },
            "type": {
                "description": "Component type",
                "examples": ["Solvent", "solvent"],
                "type": "string",
                "enum": [
                    "solvent",
                    "solute",
                    "anion",
                    "cation",
                    "Solvent",
                    "Solute",
                    "Anion",
                    "Cation",
                    "Component",
                    "component",
                ],
            },
            "valid_phase_types": {
                "type": "array",
                "items": {
                    "type": "string",
                    "description": "Valid phase types should start with 'PT.' and then match "
                    "attributes in idaes.core.base.phases.PhaseType",
                    "examples": [["PT.liquidPhase"]],
                },
            },
            "phase_equilibrium_form": {
                "type": "object",
                "properties": {
                    phase1 + "-" + phase2: {"type": "string"}
                    for phase1 in Reaction.PHASES
                    for phase2 in Reaction.PHASES
                },
            },
            "parameter_data": {
                "type": "object",
                "properties": {
                    "mw": {"$ref": "#/definitions/parameter"},
                    "pressure_crit": {"$ref": "#/definitions/parameter"},
                    "temperature_crit": {"$ref": "#/definitions/parameter"},
                },
                "patternProperties": {
                    "^.*_coeff$": {"$ref": "#/definitions/parameter"},
                    "^.*_ref$": {"$ref": "#/definitions/parameter"},
                    "^.*_comp$": {"$ref": "#/definitions/parameter"},
                },
                "additionalProperties": False,
            },
        },
        "required": ["name", "elements", "parameter_data", "type"],
        "patternProperties": {"_comp": {"type": "string"}},
        "additionalProperties": True,
        "definitions": {"parameter": _parameter_def},
    },
    "reaction": {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "$id": "https://nawi-hub.github.com/electrolytedb/reaction",
        "description": "The stoichiometry and properties of a reaction",
        "type": "object",
        "properties": {
            "type": {
                "type": "string",
                "enum": ["equilibrium", "inherent"],
                "description": "Type of reaction",
            },
            "name": {"type": "string", "description": "Name of reaction"},
            "compnents": {
                "type": "array",
                "items": {
                    "type": "string",
                    "description": "Name of a component of the reaction",
                },
            },
            Reaction.NAMES.stoich: {
                "type": "object",
                "description": "Moles for the given species in the reaction. "
                "Grouped by phase.",
                "properties": {
                    phase: {"$ref": "#/definitions/stoichiometry"}
                    for phase in Reaction.PHASES
                },
                "additionalProperties": False,
            },
            Reaction.NAMES.hor: {"type": "string"},
            Reaction.NAMES.eq_const: {"type": "string"},
            Reaction.NAMES.eq_form: {"type": "string"},
            Reaction.NAMES.conc_form: {"type": "string"},
            Reaction.NAMES.param: {
                "type": "object",
                "patternProperties": {
                    "_ref": {"$ref": "#/definitions/parameter"},
                },
                "properties": {
                    Reaction.NAMES.reaction_order: {
                        "type": "object",
                        "properties": {
                            phase: {"$ref": "#/definitions/stoichiometry"}
                            for phase in Reaction.PHASES
                        },
                        "additionalProperties": False,
                    },
                },
                "required": [Reaction.NAMES.reaction_order],
            },
        },
        "required": ["name", "parameter_data", "type", "components"],
        "definitions": {
            "parameter": _parameter_def,
            "stoichiometry": {
                "description": "One part of the stoichiometry",
                "examples": ['{"NH4 +": -1}'],
                "type": "object",
                "patternProperties": {
                    "^[A-Z].*$": {
                        "type": "number",
                        "description": "Moles for the given species in the reaction. Negative for LHS, positive for RHS",
                    }
                },
                "additionalProperties": False,
            },
            "reaction_order": {
                "description": "Reaction order for a given phase",
                "examples": ['{"H2O": 0, "H +": 1, "OH -": 1}'],
                "type": "object",
                "patternProperties": {
                    "^[A-Z].*$": {
                        "type": "number",
                        "description": "Reaction side: 0 for LHS, 1 for RHS",
                    }
                },
                "additionalProperties": False,
            },
        },
    },
}



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pathlib

_ROOT = pathlib.Path(__file__).parent.absolute()

# create the ipopt-watertap solver and register
# it as the default IDAES solver
import watertap.core.plugins



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest
from watertap.examples.flowsheets.full_treatment_train.analysis.multi_sweep import *

pytest_parameterize_list = []
# We are skipping cases 5 & 7 in order to keep the pytests times in check
for case_num in [1, 2, 3, 4, 6, 8, 9]:
    for ro_type in ["0D", "1D"]:
        test_case = (case_num, ro_type)
        pytest_parameterize_list.append(test_case)


@pytest.mark.requires_idaes_solver
@pytest.mark.parametrize("case_num, RO_type", pytest_parameterize_list)
@pytest.mark.integration
def test_multi_sweep(case_num, RO_type, tmp_path):
    nx = 1
    global_results, sweep_params = run_analysis(
        case_num, nx, RO_type, output_directory=str(tmp_path), interp_nan_outputs=False
    )


@pytest.mark.unit
def test_get_output_directory():
    path_list = ["/mypath/", None]
    return_list = [get_output_directory(dirpath) for dirpath in path_list]
    assert return_list == ["/mypath/", "./"]



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains methods for constructing the material balances for
zero-order single inlet-double outlet (SIDO) unit models with chemical
reactions.
"""
from types import MethodType

import idaes.logger as idaeslog
from idaes.core.solvers import get_solver
import idaes.core.util.scaling as iscale
from idaes.core.util.exceptions import InitializationError

from pyomo.environ import (
    check_optimal_termination,
    NonNegativeReals,
    Set,
    Var,
    units as pyunits,
)

# Some more information about this module
__author__ = "Andrew Lee"

# Set up logger
_log = idaeslog.getLogger(__name__)


def build_sido_reactive(self):
    """
    Helper method for constructing material balances for zero-order type models
    with one inlet and two outlets including chemical reactions.

    Three StateBlocks are added with corresponding Ports:
        * properties_in
        * properties_treated
        * properties_byproduct

    Additional variables are:
        * recovery_vol (indexed by time)
        * removal_frac_mass_comp (indexed by time and component)
        * extent_of_reaction (indexed by time and reactions)

    Four additional constraints are added to represent the material balances
        * water_recovery_equation (indexed by time)
        * flow_balance (indexed by time)
        * solute_removal_equation (indexed by time and solute)
        * solute_treated_equation (indexed by time and solute)

    This method also sets private attributes on the unit model with references
    to the appropriate initialization and scaling methods to use and to return
    the inlet volumetric flow rate.
    """
    self._has_recovery_removal = True
    self._initialize = MethodType(initialize_sidor, self)
    self._scaling = MethodType(calculate_scaling_factors_sidor, self)

    # Create state blocks for inlet and outlets
    tmp_dict = dict(**self.config.property_package_args)
    tmp_dict["has_phase_equilibrium"] = False
    tmp_dict["defined_state"] = True

    self.properties_in = self.config.property_package.build_state_block(
        self.flowsheet().time, doc="Material properties at inlet", **tmp_dict
    )

    tmp_dict_2 = dict(**tmp_dict)
    tmp_dict_2["defined_state"] = False

    self.properties_treated = self.config.property_package.build_state_block(
        self.flowsheet().time, doc="Material properties of treated water", **tmp_dict_2
    )
    self.properties_byproduct = self.config.property_package.build_state_block(
        self.flowsheet().time,
        doc="Material properties of byproduct stream",
        **tmp_dict_2,
    )

    # Create Ports
    self.add_port("inlet", self.properties_in, doc="Inlet port")
    self.add_port("treated", self.properties_treated, doc="Treated water outlet port")
    self.add_port("byproduct", self.properties_byproduct, doc="Byproduct outlet port")

    # Load reaction identifiers from database
    dbparams = self.config.database.get_unit_operation_parameters(
        self._tech_type, subtype=self.config.process_subtype
    )
    try:
        rxn_ids = list(dbparams["reactions"].keys())
    except KeyError:
        raise KeyError(
            f"{self.name} - database provided does not contain a list of "
            f"reactions for this technology. "
            f"Tech type: {self._tech_type}, Process subtype: {self.config.process_subtype}"
            f"Database Parameters: {dbparams}\n"
            f"Datbase path: {self.config.database._dbpath}"
            f"END"
        )
    # Create indexing set for reactions
    self.reaction_set = Set(initialize=rxn_ids)

    # Add performance variables
    self.recovery_frac_mass_H2O = Var(
        self.flowsheet().time,
        initialize=0.8,
        domain=NonNegativeReals,
        units=pyunits.dimensionless,
        bounds=(0.0, 1.0000001),
        doc="Mass recovery fraction of water in the treated stream",
    )
    self.removal_frac_mass_comp = Var(
        self.flowsheet().time,
        self.config.property_package.solute_set,
        domain=NonNegativeReals,
        initialize=0.01,
        units=pyunits.dimensionless,
        doc="Solute removal fraction on a mass basis",
    )

    self.reaction_conversion = Var(
        self.flowsheet().time,
        self.reaction_set,
        initialize=0.5,
        units=pyunits.dimensionless,
        doc="Reaction conversion based on key reactant",
    )

    # Load reaction conversions
    for (t, r), v in self.reaction_conversion.items():
        try:
            conv = dbparams["reactions"][r]["conversion"]
        except KeyError:
            raise KeyError(
                f"{self.name} - database provided does not "
                f"contain an entry for conversion for reaction {r}."
            )
        v.fix(conv)

    # Intermediate variables and parameters
    self.extent_of_reaction = Var(
        self.flowsheet().time,
        self.reaction_set,
        initialize=0,
        units=pyunits.kg / pyunits.s,
        doc="Reaction extent on a mass basis based on key reactant",
    )

    # Reaction extent equation
    @self.Constraint(
        self.flowsheet().time,
        self.reaction_set,
        doc="Calculation of reaction extent from conversion",
    )
    def reaction_extent_equation(b, t, r):
        # Get key reactant from database
        try:
            key_reactant = dbparams["reactions"][r]["key_reactant"]
        except KeyError:
            raise KeyError(
                f"{self.name} - database provided does not "
                f"contain an entry for key_reactant for reaction {r}."
            )
        if key_reactant not in b.properties_in[0].component_list:
            raise ValueError(
                f"{self.name} - key_reactant {key_reactant} for reaction {r} "
                f"is not in the component list used by the assigned property "
                f"package."
            )
        return (
            b.extent_of_reaction[t, r]
            == b.reaction_conversion[t, r]
            * b.properties_in[t].flow_mass_comp[key_reactant]
        )

    self.generation_ratio = Var(
        self.reaction_set,
        self.config.property_package.component_list,
        initialize=0,
        units=pyunits.dimensionless,
        doc="Mass ratio for generation of species w.r.t. key species",
    )

    # Load generation ratio data
    for (r, j), p in self.generation_ratio.items():
        # Check to see whether to use stoichiometry or conversion ratio
        try:
            stoich = dbparams["reactions"][r]["stoichiometry"]
        except KeyError:
            raise KeyError(
                f"{self.name} - database provided does not "
                f"contain an entry for stoichiometry for reaction {r}."
            )
        if j in stoich:
            if (
                "conversion_ratio" in stoich[j].keys()
                and "order" not in stoich[j].keys()
            ):
                cratio = stoich[j]["conversion_ratio"]
            elif "conversion_ratio" in stoich[j].keys() and "order" in stoich[j].keys():
                raise RuntimeError(
                    f"{self.name} - database provides entries for both "
                    f"conversion_ratio and reaction order in reaction {r}. "
                    f"Please provide only one or the other."
                )
            elif "order" in stoich[j].keys():
                key_reactant = dbparams["reactions"][r]["key_reactant"]
                if j == key_reactant:
                    cratio = -1
                else:
                    nu_j = stoich[j]["order"]
                    try:
                        nu_k = stoich[key_reactant]["order"]
                    except KeyError:
                        raise KeyError(
                            f"{self.name} - database provided does not "
                            f"contain an entry for order w.r.t. species "
                            f"{key_reactant} in reaction {r}."
                        )
                    try:
                        MW_j = stoich[j]["molecular_weight"]
                    except KeyError:
                        raise KeyError(
                            f"{self.name} - database provided does not "
                            f"contain an entry for molecular_weight w.r.t. "
                            f"species {j} in reaction {r}."
                        )
                    try:
                        MW_k = stoich[key_reactant]["molecular_weight"]
                    except KeyError:
                        raise KeyError(
                            f"{self.name} - database provided does not "
                            f"contain an entry for molecular_weight w.r.t. "
                            f"species {key_reactant} in reaction {r}."
                        )
                    cratio = -(nu_j * MW_j) / (nu_k * MW_k)
            else:
                raise RuntimeError(
                    f"{self.name} - database provided does not "
                    f"contain any information for conversion_ratio or reaction "
                    f"order w.r.t. species {j} in reaction {r}."
                )
            p.fix(cratio)
        else:
            p.fix(0)

    # Add Expression for generation of each species
    @self.Expression(
        self.flowsheet().time,
        self.reaction_set,
        self.config.property_package.component_list,
        doc="Water recovery equation",
    )
    def generation_rxn_comp(b, t, r, j):
        return b.generation_ratio[r, j] * b.extent_of_reaction[t, r]

    # Add performance constraints
    # Water recovery
    @self.Constraint(self.flowsheet().time, doc="Water recovery equation")
    def water_recovery_equation(b, t):
        return (
            b.recovery_frac_mass_H2O[t]
            * (
                b.properties_in[t].flow_mass_comp["H2O"]
                + sum(b.generation_rxn_comp[t, r, "H2O"] for r in self.reaction_set)
            )
            == b.properties_treated[t].flow_mass_comp["H2O"]
        )

    # Flow balance
    @self.Constraint(self.flowsheet().time, doc="Overall flow balance")
    def water_balance(b, t):
        return (
            b.properties_in[t].flow_mass_comp["H2O"]
            + sum(b.generation_rxn_comp[t, r, "H2O"] for r in self.reaction_set)
            == b.properties_treated[t].flow_mass_comp["H2O"]
            + b.properties_byproduct[t].flow_mass_comp["H2O"]
        )

    # Solute removal
    @self.Constraint(
        self.flowsheet().time,
        self.config.property_package.solute_set,
        doc="Solute removal equations",
    )
    def solute_removal_equation(b, t, j):
        return (
            b.removal_frac_mass_comp[t, j]
            * (
                b.properties_in[t].flow_mass_comp[j]
                + sum(b.generation_rxn_comp[t, r, j] for r in self.reaction_set)
            )
            == b.properties_byproduct[t].flow_mass_comp[j]
        )

    # Solute concentration of treated stream
    @self.Constraint(
        self.flowsheet().time,
        self.config.property_package.solute_set,
        doc="Constraint for solute concentration in treated " "stream.",
    )
    def solute_treated_equation(b, t, j):
        return (
            b.properties_in[t].flow_mass_comp[j]
            + sum(b.generation_rxn_comp[t, r, j] for r in self.reaction_set)
            == b.properties_treated[t].flow_mass_comp[j]
            + b.properties_byproduct[t].flow_mass_comp[j]
        )

    self._stream_table_dict = {
        "Inlet": self.inlet,
        "Treated": self.treated,
        "Byproduct": self.byproduct,
    }

    self._perf_var_dict["Water Recovery"] = self.recovery_frac_mass_H2O
    self._perf_var_dict["Solute Removal"] = self.removal_frac_mass_comp
    self._perf_var_dict["Reaction Extent"] = self.extent_of_reaction

    self._get_Q = MethodType(_get_Q_sidor, self)


def initialize_sidor(
    blk, state_args=None, outlvl=idaeslog.NOTSET, solver=None, optarg=None
):
    """
    Initialization routine for single inlet-double outlet unit models with
    reactions.

    Keyword Arguments:
        state_args : a dict of arguments to be passed to the property
                        package(s) to provide an initial state for
                        initialization (see documentation of the specific
                        property package) (default = {}).
        outlvl : sets output level of initialization routine
        optarg : solver options dictionary object (default=None, use
                  default solver options)
        solver : str indicating which solver to use during
                  initialization (default = None, use default IDAES solver)

    Returns:
        None
    """
    if optarg is None:
        optarg = {}

    # Set solver options
    init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
    solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

    solver_obj = get_solver(solver, optarg)

    # Get initial guesses for inlet if none provided
    if state_args is None:
        state_args = {}
        state_dict = blk.properties_in[
            blk.flowsheet().time.first()
        ].define_port_members()

        for k in state_dict.keys():
            if state_dict[k].is_indexed():
                state_args[k] = {}
                for m in state_dict[k].keys():
                    state_args[k][m] = state_dict[k][m].value
            else:
                state_args[k] = state_dict[k].value

    # ---------------------------------------------------------------------
    # Initialize control volume block
    flags = blk.properties_in.initialize(
        outlvl=outlvl,
        optarg=optarg,
        solver=solver,
        state_args=state_args,
        hold_state=True,
    )
    blk.properties_treated.initialize(
        outlvl=outlvl,
        optarg=optarg,
        solver=solver,
        state_args=state_args,
        hold_state=False,
    )
    blk.properties_byproduct.initialize(
        outlvl=outlvl,
        optarg=optarg,
        solver=solver,
        state_args=state_args,
        hold_state=False,
    )

    init_log.info_high("Initialization Step 1 Complete.")

    # ---------------------------------------------------------------------
    # Solve unit
    with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
        results = solver_obj.solve(blk, tee=slc.tee)

    init_log.info_high("Initialization Step 2 {}.".format(idaeslog.condition(results)))

    # ---------------------------------------------------------------------
    # Release Inlet state
    blk.properties_in.release_state(flags, outlvl)

    init_log.info("Initialization Complete: {}".format(idaeslog.condition(results)))

    if not check_optimal_termination(results):
        raise InitializationError(
            f"{blk.name} failed to initialize successfully. Please check "
            f"the output logs for more information."
        )


def calculate_scaling_factors_sidor(self):
    # Get default scale factors and do calculations from base classes
    for t, v in self.water_recovery_equation.items():
        iscale.constraint_scaling_transform(
            v,
            iscale.get_scaling_factor(
                self.properties_in[t].flow_mass_comp["H2O"],
                default=1,
                warning=True,
                hint=" for water recovery",
            ),
        )

    for t, v in self.water_balance.items():
        iscale.constraint_scaling_transform(
            v,
            iscale.get_scaling_factor(
                self.properties_in[t].flow_mass_comp["H2O"], default=1, warning=False
            ),
        )  # would just be a duplicate of above

    for (t, j), v in self.solute_removal_equation.items():
        iscale.constraint_scaling_transform(
            v,
            iscale.get_scaling_factor(
                self.properties_in[t].flow_mass_comp[j],
                default=1,
                warning=True,
                hint=" for solute removal",
            ),
        )

    for (t, j), v in self.solute_treated_equation.items():
        iscale.constraint_scaling_transform(
            v,
            iscale.get_scaling_factor(
                self.properties_in[t].flow_mass_comp[j], default=1, warning=False
            ),
        )  # would just be a duplicate of above

    dbparams = self.config.database.get_unit_operation_parameters(
        self._tech_type, subtype=self.config.process_subtype
    )
    for (t, r), v in self.reaction_extent_equation.items():
        key_reactant = dbparams["reactions"][r]["key_reactant"]
        iscale.constraint_scaling_transform(
            v,
            iscale.get_scaling_factor(
                self.properties_in[t].flow_mass_comp[key_reactant],
                default=1,
                warning=False,
            ),
        )  # would just be a duplicate of above


def _get_Q_sidor(self, t):
    return self.properties_in[t].flow_vol



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import watertap.examples.flowsheets.full_treatment_train.analysis.flowsheet_NF as flowsheet_NF


# def build_components(m):
#     flowsheet_NF.build_components(m, has_bypass=False)
#
#
# def build(m):
#     flowsheet_NF.build(m, has_bypass=False)
#
#
# def scale(m):
#     flowsheet_NF.scale(m, has_bypass=False)
#
#
# def initialize(m):
#     flowsheet_NF.initialize(m, has_bypass=False)
#
#
# def report(m):
#     flowsheet_NF.report(m, has_bypass=False)


def solve_flowsheet():
    m = flowsheet_NF.solve_flowsheet(has_bypass=False)
    return m


def simulate(m, **kwargs):
    if kwargs is not None:
        if kwargs["unfix_nf_area"]:
            m.fs.NF.area.unfix()
        if "check_termination" in kwargs.keys():
            check_termination = kwargs["check_termination"]
        else:
            check_termination = True  # Defaults to True

    return flowsheet_NF.simulate(m, check_termination=check_termination)
    # return m


if __name__ == "__main__":
    solve_flowsheet()



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains a zero-order representation of a brine concentrator unit.
"""

import pyomo.environ as pyo
from pyomo.environ import units as pyunits, Var
from idaes.core import declare_process_block_class

from watertap.core import build_sido, ZeroOrderBaseData

# Some more information about this module
__author__ = "Adam Atia"


@declare_process_block_class("BrineConcentratorZO")
class BrineConcentratorZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a brine concentrator unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "brine_concentrator"

        build_sido(self)

        if "tds" not in self.config.property_package.solute_set:
            raise KeyError(
                "TDS must be included in the solute list for determining"
                " electricity intensity and power consumption of the brine "
                "concentrator unit."
            )

        # Fitting parameters based on regressions for capital and electricity
        # developed from data in Table 5.1, Table A2.3 in:
        # Survey of High-Recovery and Zero Liquid Discharge Technologies for
        # Water Utilities (2008). WateReuse Foundation:
        # https://www.waterboards.ca.gov/water_issues/programs/grants_loans/water_recycling/research/02_006a_01.pdf
        # Capital = f(TDS, recovery, flow)
        # Electricity = f(TDS, recovery, flow)
        self.elec_coeff_1 = Var(
            units=pyunits.kWh / pyunits.m**3,
            doc="Constant 1 in electricity intensity equation",
        )
        self.elec_coeff_2 = Var(
            units=pyunits.L / pyunits.mg * pyunits.kWh / pyunits.m**3,
            doc="Constant 2 in electricity intensity equation",
        )
        self.elec_coeff_3 = Var(
            units=pyunits.kWh / pyunits.m**3,
            doc="Constant 3 in electricity intensity equation",
        )
        self.elec_coeff_4 = Var(
            units=pyunits.kWh / pyunits.m**6 * pyunits.hour,
            doc="Constant 4 in electricity intensity equation",
        )

        self._fixed_perf_vars.append(self.elec_coeff_1)
        self._fixed_perf_vars.append(self.elec_coeff_2)
        self._fixed_perf_vars.append(self.elec_coeff_3)
        self._fixed_perf_vars.append(self.elec_coeff_4)

        self.electricity = Var(
            self.flowsheet().config.time,
            units=pyunits.kW,
            bounds=(0, None),
            doc="Power consumption of brine concentrator",
        )
        self.electricity_intensity = Var(
            self.flowsheet().config.time,
            units=pyunits.kWh / pyunits.m**3,
            doc="Specific energy consumption with respect to feed flowrate",
        )

        @self.Constraint(
            self.flowsheet().config.time, doc="Electricity intensity constraint"
        )
        def electricity_intensity_constraint(b, t):
            q_in = pyunits.convert(
                b.properties_in[t].flow_vol, to_units=pyunits.m**3 / pyunits.hour
            )
            tds_in = pyunits.convert(
                b.properties_in[t].conc_mass_comp["tds"],
                to_units=pyunits.mg / pyunits.L,
            )
            return (
                b.electricity_intensity[t]
                == b.elec_coeff_1
                + b.elec_coeff_2 * tds_in
                + b.elec_coeff_3 * b.recovery_frac_mass_H2O[t]
                + b.elec_coeff_4 * q_in
            )

        @self.Constraint(
            self.flowsheet().config.time, doc="Power consumption constraint"
        )
        def electricity_constraint(b, t):
            q_in = pyunits.convert(
                b.properties_in[t].flow_vol, to_units=pyunits.m**3 / pyunits.hour
            )
            return b.electricity[t] == b.electricity_intensity[t] * q_in

        self._perf_var_dict["Power Consumption (kW)"] = self.electricity
        self._perf_var_dict[
            "Electricity intensity per Inlet Flowrate  (kWh/m3)"
        ] = self.electricity_intensity

    @property
    def default_costing_method(self):
        return self.cost_brine_concentrator

    @staticmethod
    def cost_brine_concentrator(blk):
        """
        General method for costing brine concentration processes. Capital cost
        is based on the volumetirc flowrate and TDS of the incoming stream and
        the water recovery.
        This method also registers the electricity demand as a costed flow.
        """
        t0 = blk.flowsheet().time.first()
        inlet_state = blk.unit_model.properties_in[t0]

        # Get parameter dict from database
        parameter_dict = blk.unit_model.config.database.get_unit_operation_parameters(
            blk.unit_model._tech_type, subtype=blk.unit_model.config.process_subtype
        )

        # Get costing parameter sub-block for this technology
        A, B, C, D = blk.unit_model._get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            [
                "capital_a_parameter",
                "capital_b_parameter",
                "capital_c_parameter",
                "capital_d_parameter",
            ],
        )

        # Determine if a costing factor is required
        factor = parameter_dict["capital_cost"]["cost_factor"]

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation",
        )

        expr = (
            pyo.units.convert(
                A, to_units=blk.config.flowsheet_costing_block.base_currency
            )
            + pyo.units.convert(
                B * inlet_state.conc_mass_comp["tds"],
                to_units=blk.config.flowsheet_costing_block.base_currency,
            )
            + pyo.units.convert(
                C * blk.unit_model.recovery_frac_mass_H2O[t0],
                to_units=blk.config.flowsheet_costing_block.base_currency,
            )
            + pyo.units.convert(
                D * inlet_state.flow_vol,
                to_units=blk.config.flowsheet_costing_block.base_currency,
            )
        )

        blk.costing_package.add_cost_factor(
            blk, parameter_dict["capital_cost"]["cost_factor"]
        )

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == blk.cost_factor * expr
        )

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity"
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for zero-order a WWTP with secondary treatment
"""
import pytest


from pyomo.environ import (
    Block,
    ConcreteModel,
    Constraint,
    value,
    Var,
    assert_optimal_termination,
)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.core import UnitModelCostingBlock

from watertap.unit_models.zero_order import SecondaryTreatmentWWTPZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.costing.zero_order_costing import ZeroOrderCosting

solver = get_solver()


class TestSecondaryTreatmentWWTPZO_w_default_removal:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(solute_list=["viruses_enteric", "tss", "foo"])

        m.fs.unit = SecondaryTreatmentWWTPZO(
            property_package=m.fs.params, database=m.db
        )

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10)
        m.fs.unit.inlet.flow_mass_comp[0, "viruses_enteric"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "foo"].fix(1)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db

        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)
        assert isinstance(model.fs.unit.electricity_consumption, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("secondary_treatment_wwtp")

        model.fs.unit.load_parameters_from_database(use_default_removal=True)

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert (
            model.fs.unit.recovery_frac_mass_H2O[0].value
            == data["recovery_frac_mass_H2O"]["value"]
        )

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            assert v.fixed
            if j not in data["removal_frac_mass_comp"]:
                assert v.value == data["default_removal_frac_mass_comp"]["value"]
            else:
                assert v.value == data["removal_frac_mass_comp"][j]["value"]

        assert model.fs.unit.energy_electric_flow_vol_inlet.fixed
        assert (
            model.fs.unit.energy_electric_flow_vol_inlet.value
            == data["energy_electric_flow_vol_inlet"]["value"]
        )

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert pytest.approx(1.3e-2, rel=1e-5) == value(
            model.fs.unit.properties_in[0].flow_vol
        )
        assert pytest.approx(76.9231, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["viruses_enteric"]
        )
        assert pytest.approx(76.9231, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["tss"]
        )
        assert pytest.approx(76.9231, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["foo"]
        )
        assert pytest.approx(0.01115859, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].flow_vol
        )
        assert pytest.approx(9.042361, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["viruses_enteric"]
        )
        assert pytest.approx(5.170008, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["tss"]
        )
        assert pytest.approx(89.617057, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["foo"]
        )
        assert pytest.approx(0.0018414, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].flow_vol
        )
        assert pytest.approx(488.267142, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["viruses_enteric"]
        )
        assert pytest.approx(511.73285, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["tss"]
        )
        assert pytest.approx(2.310531993e-8, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["foo"]
        )
        assert pytest.approx(6.552, abs=1e-5) == value(model.fs.unit.electricity[0])

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(
                value(
                    model.fs.unit.inlet.flow_mass_comp[0, j]
                    - model.fs.unit.treated.flow_mass_comp[0, j]
                    - model.fs.unit.byproduct.flow_mass_comp[0, j]
                )
            )

    @pytest.mark.component
    def test_report(self, model):

        model.fs.unit.report()


db = Database()
params = db._get_technology("secondary_treatment_wwtp")


class Test_SecondaryTreatmentWWTPZOsubtype:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(solute_list=["viruses_enteric"])

        m.fs.unit = SecondaryTreatmentWWTPZO(property_package=m.fs.params, database=db)

        return m

    @pytest.mark.parametrize("subtype", [params.keys()])
    @pytest.mark.component
    def test_load_parameters(self, model, subtype):
        model.fs.unit.config.process_subtype = subtype
        data = db.get_unit_operation_parameters(
            "secondary_treatment_wwtp", subtype=subtype
        )

        model.fs.unit.load_parameters_from_database()

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            assert v.fixed
            assert v.value == data["removal_frac_mass_comp"][j]["value"]


@pytest.mark.parametrize("subtype", [k for k in params.keys()])
def test_costing(subtype):
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = WaterParameterBlock(solute_list=["sulfur", "toc", "tss"])

    m.fs.costing = ZeroOrderCosting()

    m.fs.unit1 = SecondaryTreatmentWWTPZO(
        property_package=m.fs.params, database=m.db, process_subtype=subtype
    )

    m.fs.unit1.inlet.flow_mass_comp[0, "H2O"].fix(10000)
    m.fs.unit1.inlet.flow_mass_comp[0, "sulfur"].fix(1)
    m.fs.unit1.inlet.flow_mass_comp[0, "toc"].fix(2)
    m.fs.unit1.inlet.flow_mass_comp[0, "tss"].fix(3)
    m.fs.unit1.load_parameters_from_database(use_default_removal=True)
    assert degrees_of_freedom(m.fs.unit1) == 0

    m.fs.unit1.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    assert isinstance(m.fs.costing.secondary_treatment_wwtp, Block)
    assert isinstance(m.fs.costing.secondary_treatment_wwtp.capital_a_parameter, Var)
    assert isinstance(m.fs.costing.secondary_treatment_wwtp.capital_b_parameter, Var)
    assert isinstance(m.fs.costing.secondary_treatment_wwtp.reference_state, Var)

    assert isinstance(m.fs.unit1.costing.capital_cost, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost_constraint, Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit1) == 0

    assert m.fs.unit1.electricity[0] in m.fs.costing._registered_flows["electricity"]



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest

from pyomo.environ import ConcreteModel, assert_optimal_termination, value
from pyomo.util.check_units import assert_units_consistent
from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.models.unit_models.heat_exchanger import (
    HeatExchanger,
    HeatExchangerFlowPattern,
)
import idaes.core.util.scaling as iscale

from watertap.unit_models.mvc.components.lmtd_chen_callback import (
    delta_temperature_chen_callback,
)
import watertap.property_models.seawater_prop_pack as props

solver = get_solver()

# -----------------------------------------------------------------------------
@pytest.mark.requires_idaes_solver
@pytest.mark.component
def test_heat_exchanger():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = props.SeawaterParameterBlock()
    m.fs.unit = HeatExchanger(
        hot_side_name="hot",
        cold_side_name="cold",
        hot={"property_package": m.fs.properties},
        cold={"property_package": m.fs.properties},
        delta_temperature_callback=delta_temperature_chen_callback,
        flow_pattern=HeatExchangerFlowPattern.countercurrent,
    )

    # scaling
    m.fs.properties.set_default_scaling("flow_mass_phase_comp", 1, index=("Liq", "H2O"))
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e2, index=("Liq", "TDS")
    )
    iscale.set_scaling_factor(m.fs.unit.hot.heat, 1e-3)
    iscale.set_scaling_factor(m.fs.unit.cold.heat, 1e-3)
    iscale.set_scaling_factor(m.fs.unit.overall_heat_transfer_coefficient, 1e-3)
    iscale.set_scaling_factor(m.fs.unit.area, 1)
    iscale.calculate_scaling_factors(m)

    # ---specifications---
    # state variables
    m.fs.unit.hot_inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(1)
    m.fs.unit.hot_inlet.flow_mass_phase_comp[0, "Liq", "TDS"].fix(0.01)
    m.fs.unit.hot_inlet.temperature[0].fix(350)
    m.fs.unit.hot_inlet.pressure[0].fix(2e5)

    m.fs.unit.cold_inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(0.5)
    m.fs.unit.cold_inlet.flow_mass_phase_comp[0, "Liq", "TDS"].fix(0.01)
    m.fs.unit.cold_inlet.temperature[0].fix(298)
    m.fs.unit.cold_inlet.pressure[0].fix(2e5)

    m.fs.unit.area.fix(5)
    m.fs.unit.overall_heat_transfer_coefficient.fix(1000)

    # solving
    assert_units_consistent(m)
    degrees_of_freedom(m)

    m.fs.unit.initialize()

    solver = get_solver()
    results = solver.solve(m, tee=False)
    assert_optimal_termination(results)

    assert pytest.approx(88038.104, rel=1e-4) == value(m.fs.unit.heat_duty[0])
    assert pytest.approx(1.0, rel=1e-4) == value(
        m.fs.unit.hot_outlet.flow_mass_phase_comp[0, "Liq", "H2O"]
    )
    assert pytest.approx(0.01, rel=1e-4) == value(
        m.fs.unit.hot_outlet.flow_mass_phase_comp[0, "Liq", "TDS"]
    )
    assert pytest.approx(328.668, rel=1e-4) == value(
        m.fs.unit.hot_outlet.temperature[0]
    )
    assert pytest.approx(2.0e5, rel=1e-4) == value(m.fs.unit.hot_outlet.pressure[0])
    assert pytest.approx(0.5, rel=1e-4) == value(
        m.fs.unit.cold_outlet.flow_mass_phase_comp[0, "Liq", "H2O"]
    )
    assert pytest.approx(0.01, rel=1e-4) == value(
        m.fs.unit.cold_outlet.flow_mass_phase_comp[0, "Liq", "TDS"]
    )
    assert pytest.approx(341.021, rel=1e-4) == value(
        m.fs.unit.cold_outlet.temperature[0]
    )
    assert pytest.approx(2.0e5, rel=1e-4) == value(m.fs.unit.cold_outlet.pressure[0])

    m.fs.unit.report()



###############################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
# OLI Systems, Inc. Copyright © 2022, all rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# 3. Neither the name of OLI Systems, Inc. nor the names of any contributors to
# the software made available herein may be used to endorse or promote products derived
# from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
# SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
# OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# You are under no obligation whatsoever to provide any bug fixes, patches, or upgrades to the
# features, functionality or performance of the source code ("Enhancements") to anyone; however,
# if you choose to make your Enhancements available either publicly, or directly to OLI Systems, Inc.,
# without imposing a separate written license agreement for such Enhancements, then you hereby grant
# the following license: a non-exclusive, royalty-free perpetual license to install, use, modify, prepare
# derivative works, incorporate into other computer software, distribute, and sublicense such enhancements
# or derivative works thereof, in binary and source code form.
###############################################################################

import contextlib
import os
from pathlib import Path

import pytest

from watertap.tools.oli_api.client import OLIApi
from watertap.tools.oli_api.credentials import (
    CredentialManager,
    cryptography_available,
)


@pytest.fixture(scope="session")
def local_dbs_file() -> Path:
    test_dir = Path(__file__).parent / "tests"
    dbs_file_path = test_dir / "test.dbs"
    return dbs_file_path


@pytest.fixture(scope="session")
def auth_credentials() -> dict:
    "Credentials that allow running tests with an authenticated client"
    creds = {"auth_url": "not required when using access keys"}
    try:
        creds["access_keys"] = [os.environ["OLI_API_KEY"]]
        creds["root_url"] = os.environ["OLI_API_ROOT_URL"]
    except KeyError as e:
        pytest.skip(f"Authenticated credentials not found in environment variable: {e}")
    return creds


@pytest.fixture(scope="function")
def oliapi_instance(
    tmp_path: Path, auth_credentials: dict, local_dbs_file: Path
) -> OLIApi:

    if not cryptography_available:
        pytest.skip(reason="cryptography module not available.")
    cred_file_path = tmp_path / "pytest-credentials.txt"

    credentials = {
        **auth_credentials,
        "config_file": cred_file_path,
    }
    credential_manager = CredentialManager(**credentials, test=True)
    credential_manager.login()
    with OLIApi(credential_manager, test=True) as oliapi:
        oliapi.get_dbs_file_id(str(local_dbs_file))
        yield oliapi
    with contextlib.suppress(FileNotFoundError):
        cred_file_path.unlink()



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
from watertap.tools.parameter_sweep import LinearSample, parameter_sweep
import watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.amo_1690.amo_1690 as amo_1690


def set_up_sensitivity(m):
    outputs = {}
    optimize_kwargs = {"fail_flag": False}
    opt_function = amo_1690.solve

    # create outputs
    outputs["LCOT"] = m.fs.costing.LCOT
    outputs["LCOT_with_revenue"] = m.fs.costing.LCOT_with_revenue
    outputs["LC_biogas"] = m.fs.costing.LC_biogas
    outputs["LC_biogas_with_revenue"] = m.fs.costing.LC_biogas_with_revenue
    outputs["LC_fertilizer"] = m.fs.costing.LC_fertilizer
    outputs["LC_fertilizer_with_revenue"] = m.fs.costing.LC_fertilizer_with_revenue

    return outputs, optimize_kwargs, opt_function


def run_analysis(case_num=1, nx=11, interpolate_nan_outputs=True, save_path=None):
    m = amo_1690.main()[0]
    outputs, optimize_kwargs, opt_function = set_up_sensitivity(m)
    sweep_params = {}

    if case_num == 1:
        sweep_params["electricity_cost"] = LinearSample(
            m.fs.costing.electricity_cost, 0.05, 0.1, nx
        )
    elif case_num == 2:
        sweep_params["biogas_cost"] = LinearSample(m.fs.costing.biogas_cost, 0.1, 1, nx)
    elif case_num == 3:
        sweep_params["fertilizer_cost"] = LinearSample(
            m.fs.costing.fertilizer_cost, 0.1, 1, nx
        )
    else:
        raise ValueError(f"{case_num} is not yet implemented")

    global_results = parameter_sweep(
        m,
        sweep_params,
        outputs,
        csv_results_file_name=save_path,
        optimize_function=opt_function,
        optimize_kwargs=optimize_kwargs,
        interpolate_nan_outputs=interpolate_nan_outputs,
    )

    return global_results, sweep_params, m


if __name__ == "__main__":
    results, sweep_params, m = run_analysis()



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Essential tests targeting the flowsheet interfaces registered and discoverable
through the standard mechanism (currently FlowsheetInterface.from_installed_packages()).
"""
import gc

import pytest
from pyomo.environ import assert_optimal_termination

from ..fsapi import FlowsheetInterface
from ..fsapi import ModelOption


@pytest.fixture(scope="class")
def fs_interface(request) -> FlowsheetInterface:
    fs_interface: FlowsheetInterface = request.param
    fs_exp = fs_interface.fs_exp

    markers_to_apply = []
    # with this parametrization setup, ``request`` will be a SubRequest
    # if we use request.applymarker(), the marker will be applied to the entire class
    # using _parent_request (unfortunately private API) lets us target more precisely
    # a single (test_function, param_value) combination
    request_where_marker_should_be_applied = request._parent_request

    if fs_exp.requires_idaes_solver:
        markers_to_apply.append(pytest.mark.requires_idaes_solver)

    for marker in markers_to_apply:
        request_where_marker_should_be_applied.applymarker(marker)

    return fs_interface


def pytest_generate_tests(metafunc):
    if "fs_interface" in metafunc.fixturenames:
        by_name = dict(FlowsheetInterface.from_installed_packages())
        # NOTE: with ``scope="class"``, a fresh, separate fixture instance will be created:
        # - for each standalone (i.e. module-level) test function,
        # - for each test class, once by the first test method in the class that requests it
        # as a consequence, this means that all test methods within a test class operate on the same object
        metafunc.parametrize(
            "fs_interface",
            list(by_name.values()),
            ids=list(by_name.keys()),
            scope="class",
            indirect=True,
        )


class TestFlowsheetInterface:
    def test_basic_type(self, fs_interface):
        assert isinstance(fs_interface, FlowsheetInterface)

    @pytest.fixture
    def data_post_build(self, fs_interface) -> dict:
        fs_interface.build()
        return fs_interface.dict()

    def test_build(self, data_post_build):
        data = data_post_build
        assert data
        assert "model_objects" in data

    @pytest.fixture
    def model_objects(self, data_post_build):
        return data_post_build["model_objects"]

    def test_model_objects(self, model_objects):
        assert len(model_objects) > 0

    @pytest.fixture
    def solve_results(self, fs_interface, data_post_build):
        res = fs_interface.solve()
        return res

    def test_solve(self, solve_results):
        assert solve_results
        assert_optimal_termination(solve_results)


@pytest.mark.parametrize("n_times", [2, 3], ids="{} times".format)
def test_roundtrip_with_garbage_collection(fs_interface, n_times):
    build_options = {
        "Bypass": ModelOption(
            name="bypass option",
            display_name="With Bypass",
            values_allowed=["false", "true"],
            value="false",
        ),
        "ConcentrationPolarization": ModelOption(
            name="ConcentrationPolarization",
            display_name="Concentration Polarization Type",
            values_allowed=["calculated", "none"],
            value="calculated",
        ),
    }

    for attempt in range(n_times):
        fs_interface.build(build_options=build_options)
        data = fs_interface.dict()
        fs_interface.load(data)
        gc.collect()



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
import pytest
from pyomo.environ import (
    ConcreteModel,
    Var,
    value,
)
from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from pyomo.util.check_units import assert_units_consistent
import watertap.property_models.NaCl_prop_pack as props
from watertap.examples.flowsheets.oaro.oaro_multi import (
    main,
    build,
    set_operating_conditions,
    initialize_system,
    optimize_set_up,
    solve,
    ERDtype,
)

solver = get_solver()

# -----------------------------------------------------------------------------
class TestOAROwithTurbine:
    @pytest.fixture(scope="class")
    def system_frame(self):
        m = build(number_of_stages=3, erd_type=ERDtype.pump_as_turbine)

        return m

    @pytest.mark.unit
    def test_build(self, system_frame):
        m = system_frame
        # model set up
        assert isinstance(m, ConcreteModel)
        assert isinstance(m.fs, FlowsheetBlock)
        assert isinstance(m.fs.properties, props.NaClParameterBlock)

        # additional variables
        assert isinstance(m.fs.water_recovery, Var)

    @pytest.mark.component
    def test_units(self, system_frame):
        assert_units_consistent(system_frame)

    @pytest.mark.component
    def test_set_operating_conditions(self, system_frame):
        m = system_frame
        set_operating_conditions(m)
        assert degrees_of_freedom(m) == 0

    @pytest.mark.requires_idaes_solver
    @pytest.mark.component
    def test_initialize_system(self, system_frame):
        m = system_frame
        initialize_system(
            m,
            number_of_stages=3,
            solvent_multiplier=0.5,
            solute_multiplier=0.7,
            solver=solver,
        )
        assert degrees_of_freedom(m) == 0

    @pytest.mark.requires_idaes_solver
    @pytest.mark.component
    def test_solution(self, system_frame):
        m = system_frame
        optimize_set_up(m, number_of_stages=3, water_recovery=0.5)
        solve(m, solver=solver)
        assert pytest.approx(1.318023e-3, rel=1e-5) == value(
            m.fs.product.flow_mass_phase_comp[0, "Liq", "NaCl"]
        )
        assert pytest.approx(0.5, rel=1e-5) == value(m.fs.mass_water_recovery)

    @pytest.mark.component
    def test_config_error(self, system_frame):
        with pytest.raises(Exception):
            build(number_of_stages=3, erd_type="not_a_configuration")

    @pytest.mark.requires_idaes_solver
    @pytest.mark.component
    def test_main(self):
        main(number_of_stages=3, system_recovery=0.5)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
import os
import numpy as np
import pyomo.environ as pyo
import copy

from watertap.tools.parameter_sweep.parameter_sweep import *
from watertap.tools.parameter_sweep.parameter_sweep_writer import *
from watertap.tools.parameter_sweep.tests.test_parameter_sweep import (
    _get_rank0_path,
    _read_output_h5,
    _assert_dictionary_correctness,
)

# ------------------------------------------------------------------------------


class TestParallelWriterManager:
    @pytest.fixture(scope="class")
    def model(self):
        m = pyo.ConcreteModel()
        m.fs = fs = pyo.Block()

        fs.input = pyo.Var(["a", "b"], within=pyo.UnitInterval, initialize=0.5)
        fs.output = pyo.Var(["c", "d"], within=pyo.UnitInterval, initialize=0.5)

        fs.slack = pyo.Var(["ab_slack", "cd_slack"], bounds=(0, 0), initialize=0.0)
        fs.slack_penalty = pyo.Param(
            default=1000.0, mutable=True, within=pyo.PositiveReals
        )

        fs.ab_constr = pyo.Constraint(
            expr=(fs.output["c"] + fs.slack["ab_slack"] == 2 * fs.input["a"])
        )
        fs.cd_constr = pyo.Constraint(
            expr=(fs.output["d"] + fs.slack["cd_slack"] == 3 * fs.input["b"])
        )

        fs.performance = pyo.Expression(expr=pyo.summation(fs.output))

        m.objective = pyo.Objective(
            expr=m.fs.performance - m.fs.slack_penalty * pyo.summation(m.fs.slack),
            sense=pyo.maximize,
        )
        return m

    @pytest.mark.unit
    def test_interp_nan_values(self):
        ps = ParameterSweep()
        ps_writer = ParameterSweepWriter(
            ps.parallel_manager,
            csv_results_file_name=None,
            h5_results_file_name=None,
            debugging_data_dir=None,
            interpolate_nan_outputs=True,
        )

        global_values = np.array(
            [
                [0, 0, 0],
                [0, 0, 1],
                [0, 1, 0],
                [0, 1, 1],
                [1, 0, 0],
                [1, 0, 1],
                [1, 1, 0],
                [1, 1, 1],
                [0.5, 0.5, 0.5],
                [1, 1, 1],
            ]
        )

        global_results = np.array([0, 1, 2, 3, 4, 5, 6, 7, np.nan, np.nan])[
            np.newaxis
        ].T

        global_results_clean = ps_writer._interp_nan_values(
            global_values, global_results
        )

        assert np.shape(global_results_clean)[1] == np.shape(global_results)[1]
        assert np.shape(global_results_clean)[0] == np.shape(global_results)[0]

        assert (global_results_clean[8]) == pytest.approx(np.mean(global_results[0:8]))
        assert (global_results_clean[9]) == pytest.approx(global_results[7])

    @pytest.mark.unit
    def test_h5_read_write(self, tmp_path):
        ps = ParameterSweep()

        tmp_path = _get_rank0_path(None, tmp_path)
        h5_fname = "h5_test_{0}.h5".format(ps.parallel_manager.get_rank())

        ps_writer = ParameterSweepWriter(
            ps.parallel_manager,
            csv_results_file_name=None,
            h5_results_file_name=h5_fname,
            debugging_data_dir=None,
            interpolate_nan_outputs=True,
        )

        input_dict = {
            "outputs": {
                "fs.input[a]": {
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
                },
                "fs.input[b]": {
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
                },
                "fs.output[c]": {
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array([0.2, 0.2, 0.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0]),
                },
                "fs.output[d]": {
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array([0.0, 0.75, 0.0, 0.0, 0.75, 0.0, 0.0, 0.0, 0.0]),
                },
                "fs.slack[ab_slack]": {
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
                },
                "fs.slack[cd_slack]": {
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
                },
            },
            "solve_successful": [True] * 9,
            "sweep_params": {
                "fs.input[a]": {
                    "lower bound": None,
                    "units": "None",
                    "upper bound": None,
                    "value": np.array([0.1, 0.1, 0.0, 0.5, 0.5, 0.0, 0.0, 0.0, 0.0]),
                },
                "fs.input[b]": {
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array([0.0, 0.25, 0.0, 0.0, 0.25, 0.0, 0.0, 0.0, 0.0]),
                },
            },
        }

        reference_dict = copy.deepcopy(input_dict)
        reference_dict["sweep_params"]["fs.input[a]"]["lower bound"] = np.finfo("d").min
        reference_dict["sweep_params"]["fs.input[a]"]["upper bound"] = np.finfo("d").max

        ps_writer._write_output_to_h5(input_dict, os.path.join(tmp_path, h5_fname))

        read_dictionary = _read_output_h5(os.path.join(tmp_path, h5_fname))
        _assert_dictionary_correctness(reference_dict, read_dictionary)

    @pytest.mark.unit
    def test_h5_parents(self, tmp_path):
        ps = ParameterSweep()

        tmp_path = _get_rank0_path(None, tmp_path)
        h5_fname = "h5_test_{0}.h5".format(ps.parallel_manager.get_rank())
        h5_parent_groups = ["loop1", "loop2"]

        embedded_dict = {
            "outputs": {
                "fs.input[a]": {
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
                },
                "fs.input[b]": {
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
                },
                "fs.output[c]": {
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array([0.2, 0.2, 0.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0]),
                },
                "fs.output[d]": {
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array([0.0, 0.75, 0.0, 0.0, 0.75, 0.0, 0.0, 0.0, 0.0]),
                },
                "fs.slack[ab_slack]": {
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
                },
                "fs.slack[cd_slack]": {
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
                },
            },
            "solve_successful": [True] * 9,
            "sweep_params": {
                "fs.input[a]": {
                    "lower bound": None,
                    "units": "None",
                    "upper bound": None,
                    "value": np.array([0.1, 0.1, 0.0, 0.5, 0.5, 0.0, 0.0, 0.0, 0.0]),
                },
                "fs.input[b]": {
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array([0.0, 0.25, 0.0, 0.0, 0.25, 0.0, 0.0, 0.0, 0.0]),
                },
            },
        }

        ps_writer_dict = dict()
        reference_dict = dict()
        for h5_parent_group in h5_parent_groups:
            ps_writer_dict[h5_parent_group] = ParameterSweepWriter(
                ps.parallel_manager,
                csv_results_file_name=None,
                h5_results_file_name=h5_fname,
                h5_parent_group_name=h5_parent_group,
                debugging_data_dir=None,
                interpolate_nan_outputs=True,
            )

            reference_dict[h5_parent_group] = embedded_dict
            reference_dict[h5_parent_group]["sweep_params"]["fs.input[a]"][
                "lower bound"
            ] = np.finfo("d").min
            reference_dict[h5_parent_group]["sweep_params"]["fs.input[a]"][
                "upper bound"
            ] = np.finfo("d").max

            ps_writer_dict[h5_parent_group]._write_output_to_h5(
                embedded_dict, os.path.join(tmp_path, h5_fname)
            )

        # Now read the dictionary
        read_dictionary = _read_loop_h5(
            os.path.join(tmp_path, h5_fname), h5_parent_groups
        )
        _assert_loop_correctness(reference_dict, read_dictionary)


def _read_loop_h5(filepath, parent_list):

    f = h5py.File(filepath, "r")

    output_dict = {}
    for parent in parent_list:
        output_dict[parent] = _read_output_h5(f[parent])

    f.close()

    return output_dict


def _assert_loop_correctness(truth_dict, test_dict):
    assert truth_dict.keys() == test_dict.keys()

    for key, item in truth_dict.items():
        _assert_dictionary_correctness(item, test_dict[key])



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Clarifier unit model for BSM2 and plant-wide wastewater treatment modeling.
This unit inherits from the IDAES separator unit.
"""

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
)
from idaes.models.unit_models.separator import SeparatorData, SplittingType

from idaes.core.util.tables import create_stream_table_dataframe
import idaes.logger as idaeslog

from pyomo.environ import (
    Var,
    Param,
    units as pyunits,
)

from idaes.core.util.exceptions import (
    ConfigurationError,
)
from watertap.costing.unit_models.clarifier import cost_clarifier

__author__ = "Chenyu Wang"

# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("Clarifier")
class ClarifierData(SeparatorData):
    """
    Thickener unit model for BSM2
    """

    CONFIG = SeparatorData.CONFIG()
    CONFIG.outlet_list = ["underflow", "overflow"]
    CONFIG.split_basis = SplittingType.componentFlow

    def build(self):
        """
        Begin building model.
        Args:
            None
        Returns:
            None
        """

        # Call UnitModel.build to set up dynamics
        super(ClarifierData, self).build()

        if "underflow" and "effluent" not in self.config.outlet_list:
            raise ConfigurationError(
                "{} encountered unrecognised "
                "outlet_list. This should not "
                "occur - please use underflow "
                "and effluent as outlets.".format(self.name)
            )

        self.surface_area = Var(
            initialize=1500,
            doc="Cross section surface area of the clarifier",
            units=pyunits.m**2,
            bounds=(0, 3000),
        )

        self.electricity_consumption = Var(
            self.flowsheet().time,
            units=pyunits.kW,
            bounds=(0, None),
            doc="Electricity consumption of unit",
        )
        # The value is taken from Maravelias' data
        self.energy_electric_flow_vol_inlet = Param(
            initialize=0.008,
            units=pyunits.kWh / pyunits.m**3,
            mutable=True,
            doc="Electricity intensity with respect to inlet flow",
        )

        # Electricity constraint
        @self.Constraint(
            self.flowsheet().time,
            doc="Constraint for electricity consumption based on phosphorus removal",
        )
        def rule_electricity_consumption(self, t):
            return self.electricity_consumption[t] == (
                self.energy_electric_flow_vol_inlet
                * pyunits.convert(
                    self.inlet.flow_vol[t],
                    to_units=pyunits.m**3 / pyunits.hr,
                )
            )

    def _get_performance_contents(self, time_point=0):
        if hasattr(self, "split_fraction"):
            var_dict = {}
            for k in self.split_fraction.keys():
                if k[0] == time_point:
                    var_dict[f"Split Fraction [{str(k[1:])}]"] = self.split_fraction[k]
            return {"vars": var_dict}
        else:
            return None

    def _get_stream_table_contents(self, time_point=0):
        outlet_list = self.create_outlet_list()

        io_dict = {}
        if self.config.mixed_state_block is None:
            io_dict["Inlet"] = self.mixed_state
        else:
            io_dict["Inlet"] = self.config.mixed_state_block

        for o in outlet_list:
            io_dict[o] = getattr(self, o + "_state")

        return create_stream_table_dataframe(io_dict, time_point=time_point)

    @property
    def default_costing_method(self):
        return cost_clarifier



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
import os
import numpy as np
import pyomo.environ as pyo
import ast

from pyomo.environ import value

from watertap.tools.parameter_sweep.sampling_types import *
from watertap.tools.parameter_sweep import (
    RecursiveParameterSweep,
    recursive_parameter_sweep,
)
from watertap.tools.parameter_sweep.parameter_sweep_writer import *
from watertap.tools.parameter_sweep.tests.test_parameter_sweep import (
    _read_output_h5,
    _get_rank0_path,
    _assert_dictionary_correctness,
)

import watertap.tools.MPI as MPI

# -----------------------------------------------------------------------------


@pytest.fixture
def model():
    """
    # Example Usage:
    # Set the number of trials
    nn = 50
    results = np.zeros((nn, 2))

    for k in range(nn):
        # Set a random value for the parameter a
        m.fs.a.set_value(np.random.fs.rand())

        # Attempt to solve the model (infeasible if m.fs.a > success_prob)
        solver.solve(m)

        # Store the value of a and the solution x
        a = pyo.value(m.fs.a)
        x = pyo.value(m.fs.x)

        if np.abs(pyo.value(m.fs.err)) < 1e-6:
            results[k, :] = [a, x]
        else:
            results[k, :] = [a, np.nan]
    """

    # Initialize a model and solver
    m = pyo.ConcreteModel()

    m.fs = pyo.Block()

    # Declare decision variable and param
    m.fs.x = pyo.Var()
    m.fs.a = pyo.Param(mutable=True)
    m.fs.success_prob = pyo.Param(initialize=0.5, mutable=True)

    # Define expressions and constraints:
    # Numbers must sum to success_prob and x must be positive
    m.fs.err = pyo.Expression(expr=m.fs.a + m.fs.x - m.fs.success_prob)
    m.fs.sum = pyo.Constraint(expr=m.fs.err == 0.0)
    m.fs.pos = pyo.Constraint(expr=m.fs.x >= 0.0)

    return m


@pytest.mark.component
def test_aggregate_filtered_input_arr():
    ps = RecursiveParameterSweep()

    input_dict = {
        "outputs": {
            "x_val": {
                "lower bound": None,
                "units": "None",
                "upper bound": None,
                "value": np.array(
                    [
                        0.31655848,
                        0.2763452,
                        0.26241279,
                        0.15511682,
                        0.1511865,
                        0.09723662,
                        0.05410589,
                        0.6797816,
                        0.62896394,
                        0.6128707,
                        0.23852064,
                        0.17110508,
                        0.13195544,
                    ]
                ),
            }
        },
        "solve_successful": [True] * 13,
        "sweep_params": {
            "fs.a": {
                "units": "None",
                "value": np.array(
                    [
                        0.38344152,
                        0.4236548,
                        0.43758721,
                        0.54488318,
                        0.5488135,
                        0.60276338,
                        0.64589411,
                        0.0202184,
                        0.07103606,
                        0.0871293,
                        0.46147936,
                        0.52889492,
                        0.56804456,
                    ]
                ),
            }
        },
    }

    truth_arr = np.array(
        [
            0.38344152,
            0.4236548,
            0.43758721,
            0.54488318,
            0.5488135,
            0.60276338,
            0.64589411,
            0.0202184,
            0.07103606,
            0.0871293,
            0.46147936,
            0.52889492,
            0.56804456,
        ]
    )

    req_num_samples = 10
    values_arr = ps._aggregate_filtered_input_arr(input_dict, req_num_samples)

    assert np.allclose(values_arr[:, 0], truth_arr[0:req_num_samples], equal_nan=True)


@pytest.mark.component
def test_recursive_parameter_sweep(model, tmp_path):
    comm = MPI.COMM_WORLD

    tmp_path = _get_rank0_path(comm, tmp_path)

    results_fname = os.path.join(tmp_path, "global_results")
    csv_results_file = str(results_fname) + ".csv"
    h5_results_file = str(results_fname) + ".h5"

    ps = RecursiveParameterSweep(
        csv_results_file_name=csv_results_file,
        h5_results_file_name=h5_results_file,
        debugging_data_dir=tmp_path,
    )

    m = model

    solver = pyo.SolverFactory("ipopt")

    num_samples = 10
    sweep_params = {}
    sweep_params["a_val"] = UniformSample(m.fs.a, 0.0, 1.0, num_samples)

    outputs = {}
    outputs["x_val"] = m.fs.x

    # Run the parameter sweep study using num_samples randomly drawn from the above range

    seed = 0

    # Run the parameter sweep
    data_array, results_dict = ps.parameter_sweep(
        m,
        sweep_params,
        build_outputs=outputs,
        num_samples=num_samples,
        seed=seed,
    )

    reference_save_data = np.array(
        [
            [0.38344152, 0.11655848],
            [0.4236548, 0.0763452],
            [0.43758721, 0.06241279],
            [0.0187898, 0.4812102],
            [0.0202184, 0.4797816],
            [0.06022547, 0.43977453],
            [0.07103606, 0.42896394],
            [0.0871293, 0.4128707],
            [0.10204481, 0.39795519],
            [0.11827443, 0.38172557],
        ]
    )

    import pprint

    # print("data = ")
    # pprint.pprint(data)
    assert np.shape(data_array) == (10, 2)
    assert np.allclose(reference_save_data, data_array, equal_nan=True)
    assert np.allclose(np.sum(data_array, axis=1), value(m.fs.success_prob))

    if ps.parallel_manager.is_root_process():
        # Check that the global results file is created
        assert os.path.isfile(csv_results_file)

        # Check that all local output files have been created
        for k in range(ps.parallel_manager.number_of_worker_processes()):
            assert os.path.isfile(os.path.join(tmp_path, f"local_results_{k:03}.h5"))
            assert os.path.isfile(os.path.join(tmp_path, f"local_results_{k:03}.csv"))

        csv_data = np.genfromtxt(csv_results_file, skip_header=1, delimiter=",")

        # Compare the last row of the imported data to truth
        assert np.allclose(
            data_array[-1, :], reference_save_data[-1, :], equal_nan=True
        )

        truth_dict = {
            "outputs": {
                "x_val": {
                    "lower bound": -1.7976931348623157e308,
                    "units": "None",
                    "upper bound": 1.7976931348623157e308,
                    "value": np.array(
                        [
                            0.11655848,
                            0.0763452,
                            0.06241279,
                            0.4812102,
                            0.4797816,
                            0.43977453,
                            0.42896394,
                            0.4128707,
                            0.39795519,
                            0.38172557,
                        ]
                    ),
                }
            },
            "solve_successful": [
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
            ],
            "sweep_params": {
                "a_val": {
                    "units": "None",
                    "value": np.array(
                        [
                            0.38344152,
                            0.4236548,
                            0.43758721,
                            0.0187898,
                            0.0202184,
                            0.06022547,
                            0.07103606,
                            0.0871293,
                            0.10204481,
                            0.11827443,
                        ]
                    ),
                }
            },
        }

        read_dict = _read_output_h5(h5_results_file)
        _assert_dictionary_correctness(truth_dict, read_dict)
        _assert_dictionary_correctness(truth_dict, results_dict)

        assert np.allclose(
            data_array[:, -1], read_dict["outputs"]["x_val"]["value"][:num_samples]
        )

        # Check for the companion text file
        txt_fpath = os.path.join(tmp_path, "{0}.txt".format(h5_results_file))
        assert os.path.exists(txt_fpath)

        truth_txt_dict = {
            "outputs": ["x_val"],
            "sweep_params": ["a_val"],
        }

        with open(txt_fpath, "r") as f:
            f_contents = f.read()
            read_txt_dict = ast.literal_eval(f_contents)
        assert read_txt_dict == truth_txt_dict


@pytest.mark.component
def test_recursive_parameter_sweep_function(model, tmp_path):
    comm = MPI.COMM_WORLD
    tmp_path = _get_rank0_path(comm, tmp_path)

    m = model

    solver = pyo.SolverFactory("ipopt")

    num_samples = 10
    sweep_params = {}
    sweep_params["a_val"] = UniformSample(m.fs.a, 0.0, 1.0, num_samples)

    outputs = {}
    outputs["x_val"] = m.fs.x

    # Run the parameter sweep study using num_samples randomly drawn from the above range
    seed = 0

    results_fname = os.path.join(tmp_path, "global_results")
    csv_results_file = str(results_fname) + ".csv"
    h5_results_file = str(results_fname) + ".h5"

    # Run the parameter sweep
    data_array, results_dict = recursive_parameter_sweep(
        m,
        sweep_params,
        build_outputs=outputs,
        csv_results_file_name=csv_results_file,
        h5_results_file_name=h5_results_file,
        num_samples=num_samples,
        debugging_data_dir=tmp_path,
        seed=seed,
    )

    reference_save_data = np.array(
        [
            [0.38344152, 0.11655848],
            [0.4236548, 0.0763452],
            [0.43758721, 0.06241279],
            [0.0187898, 0.4812102],
            [0.0202184, 0.4797816],
            [0.06022547, 0.43977453],
            [0.07103606, 0.42896394],
            [0.0871293, 0.4128707],
            [0.10204481, 0.39795519],
            [0.11827443, 0.38172557],
        ]
    )

    assert np.shape(data_array) == (10, 2)
    assert np.allclose(reference_save_data, data_array, equal_nan=True)
    assert np.allclose(np.sum(data_array, axis=1), value(m.fs.success_prob))

    if comm.rank == 0:
        # Check that the global results file is created
        assert os.path.isfile(csv_results_file)

        # Check that all local output files have been created
        for k in range(comm.size):
            assert os.path.isfile(os.path.join(tmp_path, f"local_results_{k:03}.h5"))
            assert os.path.isfile(os.path.join(tmp_path, f"local_results_{k:03}.csv"))

        csv_data = np.genfromtxt(csv_results_file, skip_header=1, delimiter=",")

        # Compare the last row of the imported data to truth
        assert np.allclose(
            data_array[-1, :], reference_save_data[-1, :], equal_nan=True
        )

        # Check for the h5 output
        truth_dict = {
            "outputs": {
                "x_val": {
                    "lower bound": -1.7976931348623157e308,
                    "units": "None",
                    "upper bound": 1.7976931348623157e308,
                    "value": np.array(
                        [
                            0.11655848,
                            0.0763452,
                            0.06241279,
                            0.4812102,
                            0.4797816,
                            0.43977453,
                            0.42896394,
                            0.4128707,
                            0.39795519,
                            0.38172557,
                        ]
                    ),
                }
            },
            "solve_successful": [
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
            ],
            "sweep_params": {
                "a_val": {
                    "units": "None",
                    "value": np.array(
                        [
                            0.38344152,
                            0.4236548,
                            0.43758721,
                            0.0187898,
                            0.0202184,
                            0.06022547,
                            0.07103606,
                            0.0871293,
                            0.10204481,
                            0.11827443,
                        ]
                    ),
                }
            },
        }

        read_dict = _read_output_h5(h5_results_file)
        _assert_dictionary_correctness(truth_dict, read_dict)
        _assert_dictionary_correctness(truth_dict, results_dict)
        assert np.allclose(
            data_array[:, -1], read_dict["outputs"]["x_val"]["value"][:num_samples]
        )

        # Check for the companion text file
        txt_fpath = os.path.join(tmp_path, "{0}.txt".format(h5_results_file))
        assert os.path.exists(txt_fpath)

        truth_txt_dict = {
            "outputs": ["x_val"],
            "sweep_params": ["a_val"],
        }

        with open(txt_fpath, "r") as f:
            f_contents = f.read()
            read_txt_dict = ast.literal_eval(f_contents)
        assert read_txt_dict == truth_txt_dict
